from typing import Iterator
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .types import COMMAND
from .utils import Command

if TYPE_CHECKING:
    from .fake_popen import FakePopen, AsyncFakePopen


class ProcessRecorder:
    """
    Recorder class that holds all FakePopen (or AsyncFakePopen) instances that were
    created by the subprocess. The class contains auxiliary
    """

    calls: List[Union["FakePopen", "AsyncFakePopen"]]

    def __init__(self) -> None:
        self.calls = []

    @property
    def first_call(self) -> Optional[Union["FakePopen", "AsyncFakePopen"]]:
        """Get the first process call"""
        if not self.calls:
            return None
        return self.calls[0]

    @property
    def last_call(self) -> Optional[Union["FakePopen", "AsyncFakePopen"]]:
        """Get the last (latest) process call"""
        if not self.calls:
            return None
        return self.calls[-1]

    def call_count(self, command: Optional[COMMAND] = None) -> int:
        """Get process call count - optionally match with arguments"""
        if not command:
            return len(self.calls)
        return len(tuple(self.get_matching_calls(command)))

    def was_called(self, command: Optional[COMMAND] = None) -> bool:
        """Check if process was called - optionally match with arguments"""
        if not self.calls:
            return False
        if not command:
            return True
        return any(self.get_matching_calls(command))

    def get_matching_calls(
        self, command: COMMAND
    ) -> Iterator[Union["FakePopen", "AsyncFakePopen"]]:
        """Get calls that match arguments"""
        if not isinstance(command, Command):
            command = Command(command)
        return (call for call in self.calls if self.calls if command == call.args)

    def clear(self) -> None:
        """Clear records"""
        self.calls.clear()



"""Main module"""
from . import exceptions
from .fake_process import FakeProcess

ProcessNotRegisteredError = exceptions.ProcessNotRegisteredError

__all__ = ["FakeProcess", "exceptions", "ProcessNotRegisteredError"]



import os
import sys
import threading
from pathlib import Path
from typing import Any as AnyType
from typing import Iterator
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

ARGUMENT = Union[str, "Any", os.PathLike]


class Thread(threading.Thread):
    """Custom thread class to capture exceptions"""

    exception: Optional[Exception] = None

    def run(self) -> None:
        try:
            super().run()
        except Exception as exc:
            self.exception = exc


class Command:
    """Command definition class."""

    __slots__ = "command"

    def __init__(
        self,
        command: Union[Sequence[ARGUMENT], str],
    ):
        if isinstance(command, str):
            command = tuple(command.split(" "))
        if not isinstance(command, (list, tuple)):
            raise TypeError("Command can be only of type string, list or tuple.")

        self.command: Tuple[ARGUMENT, ...] = tuple(
            os.fspath(c) if isinstance(c, os.PathLike) else c for c in command
        )

        for (i, command_elem) in enumerate(self.command):
            if isinstance(command_elem, Any) and isinstance(
                self._get_next_command_elem(i), Any
            ):
                raise AttributeError("Cannot use `Any()` one after another.")

    def __eq__(self, other: AnyType) -> bool:
        if isinstance(other, str):
            other = other.split(" ")

        norm_command = [
            os.fspath(c) if isinstance(c, os.PathLike) else c for c in self.command
        ]
        norm_other = [os.fspath(c) if isinstance(c, os.PathLike) else c for c in other]

        if norm_other == norm_command:
            # straightforward matching
            return True

        for (i, command_elem) in enumerate(norm_command):
            if isinstance(command_elem, Any):
                next_command_elem = self._get_next_command_elem(i)
                if next_command_elem is None:
                    if not self._are_thresholds_ok(command_elem, len(norm_other)):
                        return False
                    return True
                else:
                    next_matching_elem = self._get_next_matching_elem_index(
                        norm_other, next_command_elem
                    )
                    if next_matching_elem is None:
                        return False
                    else:
                        if not self._are_thresholds_ok(
                            command_elem, next_matching_elem
                        ):
                            return False

                        norm_other = norm_other[next_matching_elem:]
            else:
                if len(norm_other) == 0 or norm_other.pop(0) != command_elem:
                    return False

        return len(norm_other) == 0

    def __iter__(self) -> Iterator:
        return iter(self.command)

    @staticmethod
    def _are_thresholds_ok(command_elem: "Any", value: int) -> bool:
        if command_elem.max is not None and value > command_elem.max:
            return False
        if command_elem.min is not None and value < command_elem.min:
            return False
        return True

    def _get_next_command_elem(self, index: int) -> Optional[ARGUMENT]:
        try:
            return self.command[index + 1]
        except IndexError:
            return None

    @staticmethod
    def _get_next_matching_elem_index(
        other: Sequence[ARGUMENT], elem: ARGUMENT
    ) -> Optional[int]:
        return next(
            (i for i, other_elem in enumerate(other) if elem == other_elem), None
        )

    def __hash__(self) -> int:
        return hash(self.command)

    def __repr__(self) -> str:
        return str(self.command)

    def __str__(self) -> str:
        return str(self.command)


class Any:
    """Wildcard definition class."""

    def __init__(self, *, min: Optional[int] = None, max: Optional[int] = None) -> None:
        if min is not None and max is not None and min > max:
            raise AttributeError("min cannot be greater than max")
        self.min: Optional[int] = min
        self.max: Optional[int] = max

    def __str__(self) -> str:
        return f"{self.__class__.__name__} (min={self.min}, max={self.max})"

    def __repr__(self) -> str:
        return str(self)


class Program:
    """Specifies the name of the final program to be executed."""

    def __init__(self, program: str) -> None:
        self.program: str = program

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.program!r})"

    def __eq__(self, other: AnyType) -> bool:
        if isinstance(other, str):
            if Path(other).name == self.program:
                return True

            if sys.platform.startswith("win"):
                for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
                    if (
                        Path(other).name.lower()
                        == Path(self.program).with_suffix(ext).name.lower()
                    ):
                        return True
        return False

    def __hash__(self) -> int:
        return hash(self.program)



"""An example script to test subprocess."""
import sys
import time

print("Stdout line 1")
print("Stdout line 2")
if "stderr" in sys.argv:
    print("Stderr line 1", file=sys.stderr)

if "wait" in sys.argv:
    time.sleep(0.5)

if "non-zero" in sys.argv:
    sys.exit(1)

if "input" in sys.argv:
    user_input = input("Provide an input: ")
    print("Provided:", user_input)



"""Additional target for mypy type checking"""
from pytest_subprocess import FakeProcess


def test_typing() -> None:
    fp = FakeProcess()
    cmd = ["ls", "-l"]
    output = ["some", "lines", "of", "output"]
    fp.register(cmd, stdout=output)



"""ProcessDispatcher class declaration"""
import asyncio
import subprocess
import sys
import typing
from collections import deque
from copy import deepcopy
from functools import partial
from typing import Any as AnyType
from typing import Awaitable
from typing import Callable
from typing import Deque
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from . import asyncio_subprocess
from . import exceptions
from .fake_popen import AsyncFakePopen
from .fake_popen import FakePopen
from .types import COMMAND
from .utils import Command

if typing.TYPE_CHECKING:
    from .fake_process import FakeProcess


class ProcessDispatcher:
    """Main class for handling processes."""

    process_list: List["FakeProcess"] = []
    built_in_popen: Optional[Callable] = None
    built_in_async_subprocess: Optional[AnyType] = None
    _allow_unregistered: bool = False
    _cache: Dict["FakeProcess", Dict["FakeProcess", AnyType]] = dict()
    _keep_last_process: bool = False
    _pid: int = 0

    @classmethod
    def register(cls, process: "FakeProcess") -> None:
        if not cls.process_list:
            cls.built_in_popen = subprocess.Popen
            subprocess.Popen = cls.dispatch  # type: ignore

            cls.built_in_async_subprocess = asyncio.subprocess
            asyncio.create_subprocess_shell = cls.async_shell  # type: ignore
            asyncio.create_subprocess_exec = cls.async_exec  # type: ignore
            asyncio.subprocess = asyncio_subprocess

        cls._cache[process] = {
            proc: deepcopy(proc.definitions) for proc in cls.process_list
        }
        cls.process_list.append(process)

    @classmethod
    def deregister(cls, process: "FakeProcess") -> None:
        cls.process_list.remove(process)
        cache = cls._cache.pop(process)
        for proc, processes in cache.items():
            proc.definitions = processes
        if not cls.process_list:
            subprocess.Popen = cls.built_in_popen  # type: ignore
            cls.built_in_popen = None

            if cls.built_in_async_subprocess is None:
                raise exceptions.PluginInternalError

            asyncio.subprocess = cls.built_in_async_subprocess
            asyncio.create_subprocess_shell = (
                cls.built_in_async_subprocess.create_subprocess_shell
            )
            asyncio.create_subprocess_exec = (
                cls.built_in_async_subprocess.create_subprocess_exec
            )
            cls.built_in_async_subprocess = None

    @classmethod
    def dispatch(
        cls, command: COMMAND, **kwargs: Optional[Dict]
    ) -> Union[FakePopen, subprocess.Popen]:
        """This method will be used instead of the subprocess.Popen()"""
        process = cls.__dispatch(command)

        if process is None:
            if cls.built_in_popen is None:
                raise exceptions.PluginInternalError

            popen: subprocess.Popen = cls.built_in_popen(command, **kwargs)
            return popen

        result = cls._prepare_instance(FakePopen, command, kwargs, process)
        if not isinstance(result, FakePopen):
            raise exceptions.PluginInternalError
        result.run_thread()
        return result

    @classmethod
    async def async_shell(
        cls, cmd: Union[str, bytes], **kwargs: Dict
    ) -> Union[AsyncFakePopen, asyncio.subprocess.Process]:
        """Replacement of  asyncio.create_subprocess_shell()"""
        if not isinstance(cmd, (str, bytes)):
            raise ValueError("cmd must be a string")
        method = partial(
            cls.built_in_async_subprocess.create_subprocess_shell,  # type: ignore
            cmd,
            **kwargs
        )
        if isinstance(cmd, bytes):
            cmd = cmd.decode()
        return await cls._call_async(cmd, method, kwargs)

    @classmethod
    async def async_exec(
        cls, program: Union[str, bytes], *args: Union[str, bytes], **kwargs: Dict
    ) -> Union[AsyncFakePopen, asyncio.subprocess.Process]:
        """Replacement of asyncio.create_subprocess_exec()"""
        if not isinstance(program, (str, bytes)):
            raise ValueError("program must be a string")

        method = partial(
            cls.built_in_async_subprocess.create_subprocess_exec,  # type: ignore
            program,
            *args,
            **kwargs
        )
        if isinstance(program, bytes):
            program = program.decode()
        command = [
            program,
            *[arg.decode() if isinstance(arg, bytes) else arg for arg in args],
        ]
        return await cls._call_async(command, method, kwargs)

    @classmethod
    async def _call_async(
        cls,
        command: COMMAND,
        async_method: Callable,
        kwargs: Dict,
    ) -> Union[AsyncFakePopen, asyncio.subprocess.Process]:
        process = cls.__dispatch(command)

        if process is None:
            if cls.built_in_async_subprocess is None:
                raise exceptions.PluginInternalError

            async_shell: Awaitable[asyncio.subprocess.Process] = async_method()
            return await async_shell

        if sys.platform == "win32" and isinstance(
            asyncio.get_event_loop_policy().get_event_loop(), asyncio.SelectorEventLoop
        ):
            raise NotImplementedError(
                "The SelectorEventLoop doesn't support subprocess"
            )

        result = cls._prepare_instance(AsyncFakePopen, command, kwargs, process)
        if not isinstance(result, AsyncFakePopen):
            raise exceptions.PluginInternalError
        result.run_thread()
        return result

    @classmethod
    def _prepare_instance(
        cls,
        klass: Union[Type[FakePopen], Type[AsyncFakePopen]],
        command: COMMAND,
        kwargs: dict,
        process: dict,
    ) -> Union[FakePopen, AsyncFakePopen]:
        # Update the command with the actual command specified by the caller.
        # This will ensure that Command objects do not end up unexpectedly in
        # caller's objects (e.g. proc.args, CalledProcessError.cmd). Take care
        # to preserve the dict that may still be referenced when using
        # keep_last_process.
        fake_popen_kwargs = process.copy()
        fake_popen_kwargs["command"] = command
        recorder = fake_popen_kwargs.pop("recorder")

        result = klass(**fake_popen_kwargs)
        recorder.calls.append(result)
        result.pid = cls._pid
        result.configure(**kwargs)
        return result

    @classmethod
    def __dispatch(cls, command: COMMAND) -> Optional[dict]:
        command_instance, processes, process_instance = cls._get_process(command)
        if process_instance:
            process_instance.calls.append(command)
        if not processes:
            if not cls.process_list[-1]._allow_unregistered:
                raise exceptions.ProcessNotRegisteredError(
                    "The process '%s' was not registered."
                    % (
                        command
                        if isinstance(command, str)
                        else " ".join(str(item) for item in command),
                    )
                )
            else:
                return None
        process = processes.popleft()
        if not processes and process_instance is not None:
            if cls.process_list[-1]._keep_last_process:
                processes.append(process)
            elif command_instance:
                del process_instance.definitions[command_instance]
        cls._pid += 1
        if isinstance(process, bool):
            # real process will be called
            return None
        return process

    @classmethod
    def _get_process(
        cls, command: COMMAND
    ) -> Tuple[
        Optional[Command], Optional[Deque[Union[dict, bool]]], Optional["FakeProcess"]
    ]:
        for proc in reversed(cls.process_list):
            command_instance, processes = next(
                (
                    (key, value)
                    for key, value in proc.definitions.items()
                    if key == command
                ),
                (None, None),
            )
            process_instance = proc
            if processes and isinstance(processes, deque):
                return command_instance, processes, process_instance
        return None, None, None
